"""
basic for proxy index using different delegated indexes

author: kapil thangavelu <k_vertigo@objectrealms.net>

$Id: testProxyIndex.py,v 1.2 2003/05/15 16:38:38 hazmat Exp $
"""

import Zope
Zope.startup()

import unittest
import os, sys
import whrandom, string, random

from OFS.SimpleItem import SimpleItem
from DateTime import DateTime

from Products.CMFCore.tests.base.testcase import SecurityRequestTest

from Products.ZCatalog import ZCatalog, Vocabulary
from Products.ZCatalog.Catalog import Catalog, CatalogError

from Products.PluginIndexes.FieldIndex.FieldIndex import FieldIndex
from Products.PluginIndexes.TextIndex.TextIndex import TextIndex
from Products.PluginIndexes.TextIndex.Lexicon import  Lexicon
from Products.PluginIndexes.KeywordIndex.KeywordIndex import KeywordIndex
from Products.PluginIndexes.DateIndex.DateIndex import DateIndex

from Products.ZCTextIndex.ZCTextIndex import ZCTextIndex, PLexicon
from Products.ZCTextIndex.Lexicon import Splitter
from Products.ZCTextIndex.Lexicon import CaseNormalizer, StopWordRemover

# time to use when querying date index
query_time = DateTime()

# verbage taken from zctextindex unit tests
text = [
    """Here's a knocking indeed! If a
    man were porter of hell-gate, he should have
    old turning the key.  knock (that made sure
    sure there's at least one word in common).""",

    """Knock,
    knock, knock! Who's there, i' the name of
    Beelzebub? Here's a farmer, that hanged
    himself on the expectation of plenty: come in
    time; have napkins enow about you; here
    you'll sweat for't.""",
]

class TestObject(SimpleItem):

    def __init__(self, id, body=''):
        self.id = id
        self.body = body

    def SearchableText(self):
        return self.body


class ProxyIndexTests( SecurityRequestTest ):
    
    def setUp(self):
        SecurityRequestTest.setUp(self)

        catalog = ZCatalog.ZCatalog('portal_catalog')
        self.root._setObject('portal_catalog', catalog)
        self.root.portal_catalog.manage_addProduct['ZCatalog'].manage_addVocabulary('Vocabulary',
                                                                                    'test vocabulary',
                                                                                    globbing=1 )

        self.root.portal_catalog._setObject('lexicon', PLexicon('lexicon', '', Splitter(), CaseNormalizer(), StopWordRemover()))


        self.cat = self.root.portal_catalog

        self.root._setObject('test1', TestObject('test1', text[0]) )
        self.root._setObject('test2', TestObject('test2', text[1]) )

    def testProxyWFieldIndex(self):

        self.cat.manage_addProduct['ProxyIndex'].manage_addProxyIndex('field_id',
                                                                      idx_type = FieldIndex.meta_type,
                                                                      value_expr = 'python: object.getId()'
                                                                      )

        test1_ob = getattr(self.root, 'test1')
        test2_ob = getattr(self.root, 'test2')
        self.cat.catalog_object(test1_ob)
        self.cat.catalog_object(test2_ob)
        res = self.cat(field_id='test1')
        
        self.assertEqual( len(res), 1, 'field index should match only 1 object')

    def testProxyWDateIndex(self):
        
        self.cat.manage_addProduct['ProxyIndex'].manage_addProxyIndex('mod_date',
                                                                      idx_type = DateIndex.meta_type,
                                                                      value_expr = 'object/bobobase_modification_time'
                                                                      )

        test1_ob = getattr(self.root, 'test1')
        test2_ob = getattr(self.root, 'test2')
        self.cat.catalog_object(test1_ob)
        self.cat.catalog_object(test2_ob)
        
        res = self.cat(mod_date={'query':query_time,
                                 'range':'min'})

        self.assertEqual(len(res), 2, 'date index query should have found 2 objects')


    def testProxyWTextIndex(self):

        self.cat.manage_addProduct['ProxyIndex'].manage_addProxyIndex('searchable_text',
                                                                      idx_type = TextIndex.meta_type,
                                                                      value_expr = 'object/SearchableText'
                                                                      )
        
        test1_ob = getattr(self.root, 'test1')
        test2_ob = getattr(self.root, 'test2')
        self.cat.catalog_object(test1_ob)
        self.cat.catalog_object(test2_ob)
        res = self.cat(searchable_text='farmer')
        self.assertEqual( len(res), 1, "Invalid Search Results for TextIndex")
        

    def testProxyWZCTextIndex(self):
        
        self.cat.manage_addProduct['ProxyIndex'].manage_addProxyIndex('zc_text',
                                                                      idx_type = ZCTextIndex.meta_type,
                                                                      value_expr = 'object/SearchableText',
                                                                      doc_attr = 'proxy_value',
                                                                      lexicon_id = 'lexicon',
                                                                      index_type = 'Cosine Measure'
                                                                      )

        test1_ob = getattr(self.root, 'test1')
        test2_ob = getattr(self.root, 'test2')
        self.cat.catalog_object(test1_ob)
        self.cat.catalog_object(test2_ob)
        res = self.cat(zc_text='gate')
        self.assertEqual( len(res), 1, "Invalid Search Results for ZCTextIndex")

    def testSearchAll(self):

        self.cat.manage_addProduct['ProxyIndex'].manage_addProxyIndex('zc_text',
                                                                      idx_type = ZCTextIndex.meta_type,
                                                                      value_expr = 'object/SearchableText',
                                                                      doc_attr = 'proxy_value',
                                                                      lexicon_id = 'lexicon',
                                                                      index_type = 'Cosine Measure'
                                                                      )

        test1_ob = getattr(self.root, 'test1')
        test2_ob = getattr(self.root, 'test2')
        self.cat.catalog_object(test1_ob)
        self.cat.catalog_object(test2_ob)

        res = self.cat()
        self.assertEqual( len(res), 2, "invalid search all results")
        
        


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(ProxyIndexTests))
    return suite

if __name__ == '__main__':
    unittest.main()



"""
Purpose:
    A pluggable index which proxies for another index
    and utilizes tales to extract an indexable attribute
    from an object.
    
Author:
    kapil thangavelu <k_vertigo@objectrealms.net>
    Copyright @2003 kapil thangavelu
Date:
    3/4/3
    
License:
    X11, see LICENSE.txt for details

$Id: ProxyIndex.py,v 1.13 2003/09/05 16:05:52 guido Exp $    
"""

import types
from UserDict import UserDict

from AccessControl import ClassSecurityInfo
from Acquisition import Implicit, aq_inner, aq_parent, aq_base
from ComputedAttribute import ComputedAttribute
from ExtensionClass import Base
from Globals import DTMLFile
from Interface.Implements import objectImplements
from Interface import Interface
from OFS.SimpleItem import SimpleItem
from zLOG import LOG, PROBLEM

from Products.PluginIndexes.common.PluggableIndex import PluggableIndexInterface
from Products.ZCatalog.IZCatalog import IZCatalog
from Products.ZCatalog.Catalog import CatalogSearchArgumentsMap
from Expression import Expression, createExprContext

_marker = ()

DEBUG_LOG_ERRORS = 0
PROXIED_INDEX_ID='proxy_value'
PROXY_EXTRA = ('idx_type', 'value_expr', 'idx_context', 'idx_caller')
CMF_FOUND = 0

# used latter when constructing an object wrapper to determine if the
# object is already wrapped.
try:
    from Products.CMFCore.CatalogTool import IndexableObjectWrapper, ICatalogTool
    CMF_FOUND = 1    
except ImportError:
    class ICatalogTool(Interface): pass
    IndexableObjectWrapper = None

class IndexableWrapper:
    def __init__(self, ob):
        self._ob = ob
    def __getattr__(self, name):
        return getattr(self._ob, name)

class RecordStyle(UserDict):
    def __getattr__(self, key):
        return self[key]
    
manage_addProxyIndexForm = DTMLFile('ui/ProxyIndexAddForm', globals())

def manage_addProxyIndex(self, id, extra=None, REQUEST=None, RESPONSE=None, URL3=None, **kw):
    """Add a proxy index"""

    rec = RecordStyle()
    if extra is not None:
        rec.update( extra )
    rec.update( kw )

    # a little bit of foo, so we can have a catalog context during
    # index construction
    rec['idx_context']=self # factory dispatcher context
    
    return self.manage_addIndex(id,
                                'ProxyIndex',
                                extra=rec, 
                                REQUEST=REQUEST,
                                RESPONSE=RESPONSE,
                                URL1=URL3)

        
class ProxyIndex(SimpleItem):

    __implements__ = PluggableIndexInterface

    meta_type = 'ProxyIndex'

    manage_options = (
        
        {'label':'Overview',
         'action':'index_overview'},
        
        {'label':'Index',
         'action':'idx/manage_workspace'}
        
        )

    index_overview = DTMLFile('ui/ProxyIndexView', globals())

    security = ClassSecurityInfo()

    def __init__(self, id, extra, caller=None):
        self.id = id

        extra['idx_caller'] = caller # save the caller for lexicon lookup
        idx_type = extra['idx_type']
        value_expr = extra['value_expr']

        self._idx_type = idx_type
        self.idx  = getIndexForType(self, idx_type, extra)
        self._expr = Expression(value_expr)

    def getId(self):
        return self.id

    def clear(self):
        clear = self.getProxyAttribute('clear')
        return clear()

    def getExpressionText(self):
        return self._expr.text

    def getIndexType(self):
        return self._idx_type

    def keyForDocument(self):
        return self.getProxyAttribute('keyForDocument')

    keyForDocument = ComputedAttribute(keyForDocument)
    
    def index_object(self, documentId, obj, threshold=None):
        index_object = self.getProxyAttribute('index_object')
        try:
            ctx = createExprContext(obj)
            value = self._expr(ctx)
            wrapper = prepareObject(obj, value)
        except Exception, e:
            # gulp... the expression could raise many different types of errors.
            if DEBUG_LOG_ERRORS:
                LOG('ProxyIndex',PROBLEM,'%s tales error while indexing'%self.id)
            return 0
        except:
            return 0
        return index_object(documentId, wrapper, threshold)

    def unindex_object(self, documentId):
        unindex_object = self.getProxyAttribute('unindex_object')
        return unindex_object(documentId)

    def _apply_index(self, request, cid=''):
        apply_index = self.getProxyAttribute('_apply_index')

        # implicit pluggable interface contract 
        if not request.has_key(self.id):
            return None
        else:
            request.keywords[PROXIED_INDEX_ID]=request.get(self.id)
        
        res = apply_index(request, cid)
        return res
        
    def numObjects(self):
        numObjects = self.getProxyAttribute('numObjects')
        if callable(numObjects):
            return numObjects()
        else:
            return numObjects
    
    def __len__(self):
        """ len """
        return self.numObjects()

    def uniqueValues(self, name=None, withLength=0):
        uniqueValues = self.getProxyAttribute('uniqueValues')
        return uniqueValues(name, withLength)

    def getEntryForObject(self, documentId, default=_marker):
        getEntryForObject = self.getProxyAttribute('getEntryForObject')
        entry = getEntryForObject(documentId, default)
        if entry is _marker:
            return None
        return entry

    def getProxyAttribute(self, name):
        if hasattr(aq_base(self.idx), name):
            return getattr(self.idx, name)

        raise AttributeError, "%s not found"%str(name)

    def items(self):
        return self.getProxyAttribute('items')()

def prepareObject(obj, value):
    """
    prepare a wrapper for indexing using obj and value
    """

    # check if it obj is already wrapped
    # via cmf catalog mechanism
    wrapper = None
    if CMF_FOUND and isinstance(obj, IndexableObjectWrapper):
        wrapper = obj

    if wrapper is None:
        wrapper = IndexableWrapper(obj)

    setattr(wrapper, PROXIED_INDEX_ID, value)
        
    return wrapper

def getIndexTypes(self, names_only=1):
    " returns a list of plugin indexes for use in a zcatalog "

    obj = self
    found = 0

    while obj is not None:
        ifaces = objectImplements(obj)
        if IZCatalog in ifaces or ICatalogTool in ifaces:
            found = 1
            break
        newobj = aq_parent(obj)
        if newobj is obj:
            obj = None
        else:
            obj = newobj

    if not found:
        return []
    
    res = obj.all_meta_types(interfaces=(PluggableIndexInterface,))
    
    if names_only:
        return [r['name'] for r in res]

    return res

def getIndexForType(self, idx_type, extra):
    # we expect to get context from extra
    ctx = extra['idx_context']
    idx_caller = extra['idx_caller']
    
    # clean up the context of our values
    for pe in PROXY_EXTRA:
        del extra[pe]
        
    # set up key values pairs for the constructor
    keys = [k for k in extra.keys() if k.startswith('key')]
    keys.sort()
    vals = [v for v in extra.keys() if v.startswith('value')]
    vals.sort()

    pairs = zip(keys, vals)
        
    for k,v in pairs:
        extra[extra[k]]=extra[v]
        del extra[k], extra[v]

    p = None
    indexes = getIndexTypes(ctx, names_only=0)    

    for i in indexes:
        if i['name'] == idx_type:
            p = i
            break

    if p is None:
        raise ValueError("Index of type %s not found"%str(idx_type))

    klass = p['instance']

    if 'extra' in klass.__init__.func_code.co_varnames:
        index = klass(PROXIED_INDEX_ID, extra=extra, caller=aq_base(idx_caller))
    else:
        index = klass(PROXIED_INDEX_ID, aq_base(idx_caller))    

    return index



"""
$Id: __init__.py,v 1.2 2003/05/09 10:51:01 hazmat Exp $
"""

import ProxyIndex

def initialize(context):

    context.registerClass(
        ProxyIndex.ProxyIndex,
        permission = 'Add Pluggable Index',
        constructors = (ProxyIndex.manage_addProxyIndexForm,
                        ProxyIndex.manage_addProxyIndex,
                        ProxyIndex.getIndexTypes,
                        ),
        icon='www/index.gif',
        visibility=None
        )
    



##############################################################################
#
# Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
# 
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
# 
##############################################################################
""" Expressions in a web-configurable workflow.

$Id: Expression.py,v 1.2 2003/05/09 10:51:01 hazmat Exp $
"""

import Globals
from Globals import Persistent
from Acquisition import aq_inner, aq_parent
from AccessControl import getSecurityManager, ClassSecurityInfo

from Products.PageTemplates.Expressions import getEngine
from Products.PageTemplates.TALES import SafeMapping
from Products.PageTemplates.Expressions import SecureModuleImporter

class Expression (Persistent):
    text = ''
    _v_compiled = None

    security = ClassSecurityInfo()

    def __init__(self, text):
        self.text = text
        self._v_compiled = getEngine().compile(text)

    def __call__(self, econtext):
        compiled = self._v_compiled
        if compiled is None:
            compiled = self._v_compiled = getEngine().compile(self.text)
        # ?? Maybe expressions should manipulate the security
        # context stack.
        res = compiled(econtext)
        if isinstance(res, Exception):
            raise res
        return res

Globals.InitializeClass(Expression)


def createExprContext( object):
    '''
    An expression context provides names for TALES expressions.
    '''

    data = {
        'object':       object,
        'nothing':      None,
        'request':      getattr( object, 'REQUEST', None ),
        'modules':      SecureModuleImporter,
        }
    return getEngine().getContext(data)
